/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gpxe/in.h>
#include <gpxe/netdevice.h>
#include <gpxe/iscsi.h>
#include <gpxe/sf_settings.h>
#include <gpxe/sanity_checks.h>

char * sc_err_msgs[] = {
	[SC_OK]= "Ok.",
	[SC_NO_NVO] = "NVO block doesn't exist.",
	[SC_VCI_TOO_LONG] = "Vendor class Identifier is too long.",
	[SC_BAD_BOOTTYPE] = "Incorrect bootype.",
	[SC_BAD_USERID] = "Username is absent or too long.",
	[SC_BAD_PASSWD] = "Password is absent or has incorrect length.",
	[SC_BAD_RUSERID] = "Reverse username is absent or too long.",
	[SC_BAD_RPASSWD] = "Reverse password is absent or has incorrect "
	                     "length.",
	[SC_BAD_ITOR_IP] = "Initiator IP address is absent or incorrect.",
	[SC_BAD_NETMASK] = "Netmask is absent or incorrect.",
	[SC_BAD_ITOR_IQN] = "Initiator IQN is absent or incorrect.",
	[SC_BAD_ITOR_DHCP_FLAG] = "[DHCP for Initiator] flag value is not "
	                          "correct.",
	[SC_BAD_TGT_DHCP_FLAG] = "[DHCP for Target] flag value is not "
	                         "correct.",
	[SC_BAD_CHAP_FLAG] = "CHAP flag value is not correct.",
	[SC_BAD_MCHAP_FLAG] = "Mutual CHAP flag value is not correct.",
	[SC_BAD_IQN_DHCP_FLAG] = "[DHCP for Initiator IQN] flag value is not "
	                         "correct.",
	[SC_BAD_TGT_NAME] = "Target server name value is incorrect.",
	[SC_BAD_TGT_IQN] = "Target IQN value is incorrect.",
	[SC_BAD_LUN] = "LUN value is incorrect.",
	[SC_BAD_PORT] = "Target port value is incorrect.",
	[SC_BAD_BOOTIMAGE] = "Bad bios bootimage type.",
};

/**
 *  Checks common for PXE and iSCSI boot mode.
 *
 * @v dev	Network device
 * @ret rc	Status code
 */
static int checks_common(struct net_device *dev)
{
	unsigned long var;
	struct settings *settings;

	/* Get nvo block */
	if ( ( settings = netdev_nvo_settings ( dev ) ) == NULL ) {
		return -SC_NO_NVO;
	}

	/* Banner delay value should be tested here if it will have some
	 * ranges
	 */

	/* Boot skip delay value should be tested here if it will have some
	 * ranges
	 */

	/* Link-up delay value should be tested here if it will have some
	 * ranges
	 */

	/* Check that boot type setting is appropriate. */
	if ( ( fetch_uint_setting ( settings, &boot_type_setting, &var ) >= 0 ) &&
	     ( var > BOOTTYPE_DISABLED ) )
		return -SC_BAD_BOOTTYPE;

	/* Check the boot image setting */
	if ( ( fetch_uint_setting ( settings, &boot_image_setting, &var ) >= 0 ) &&
	     ( var > BOOTIMAGE_NONE ) )
		return -SC_BAD_BOOTIMAGE;

	return 0;
}


/**
 *  Checks for PXE boot mode.
 *
 * @v dev	Network device
 * @ret rc	Status code
 */
static int checks_pxe(struct net_device *dev)
{
	return checks_common(dev);
}

#ifdef BCV_PROTO_ISCSI
static int sf_check_chap (struct settings *setts, struct setting *flag_sett,
                          struct setting *user_sett,
                          struct setting *pass_sett) {
	long flag;
	char buf[16];	/* Buffer contents unused */
	int reverse, rc;

	reverse = (flag_sett == &chap_target_setting);

	/* Check that chap flag setting has value 0 or 1. */
	flag = sf_sett_get_intz ( setts, flag_sett );
	if ( flag != 0 && flag != 1 )
		return (reverse ? -SC_BAD_MCHAP_FLAG : -SC_BAD_CHAP_FLAG);

	/* Check username and password only if the flag is set. */
	if ( flag == 0 )
		return 0;

	rc = fetch_string_setting  ( setts, user_sett, buf, sizeof ( buf ) );
	if ( rc < 0 || rc > SF_USERID_LEN_MAX )
		return (reverse ? -SC_BAD_RUSERID : -SC_BAD_USERID);

	rc = fetch_string_setting  ( setts, pass_sett, buf, sizeof ( buf ) );
	if ( rc < SF_PASSWD_LEN_MIN || rc > SF_PASSWD_LEN_MAX )
		return (reverse ? -SC_BAD_RPASSWD : -SC_BAD_PASSWD);

	return 0;
}


/**
 *  Checks for iSCSI boot mode excluding "cross NIC's" checks.
 *
 * @v dev	Network device
 * @ret rc	Status code
 */
static int checks_iscsi_single(struct net_device *dev)
{
	long var;
	long dhcp;
	char buf[256];	/* Buffer contents used, to check root path parses */
	struct in_addr address;
	struct settings *settings;
	int rc;

	if ((rc = checks_common(dev)) != 0)
		return rc;

	settings = netdev_nvo_settings(dev);

	/* Check that vendor class ID setting is exists and not too long. */
	rc = fetch_string_setting  ( settings, &dhcp_vendor_class_setting,
				     buf, sizeof ( buf ) );
	if ( rc > SF_VCI_LEN_MAX )
		return -SC_VCI_TOO_LONG;

	/* Check that initiator dhcp flag has value 0 or 1. */
	dhcp = sf_sett_get_intz ( settings, &initiator_dhcp_setting );
	if ( dhcp != 0 && dhcp != 1 )
		return -SC_BAD_ITOR_DHCP_FLAG;

	/* Check initiator parameters only if value of initiator dhcp flag
	 * is zerro.
	 */
	if ( !dhcp ) {
		/* Check that YIADDR is present and is not INADDR_ANY.*/
		if ( fetch_ipv4_setting ( settings, &ip_setting,
		                          &address ) < 0 ||
		     address.s_addr == 0) {
			return -SC_BAD_ITOR_IP;
		}
		/* Check that netmask is present and is not zero. */
		if ( fetch_ipv4_setting ( settings, &netmask_setting,
		                          &address ) < 0 ||
		     address.s_addr == 0) {
			return -SC_BAD_NETMASK;
		}
		/* Check that length of initiator iQN setting is not too
		 * long.
		 */
		rc = fetch_string_setting  ( settings,
					     &initiator_iqn_setting,
					     buf, sizeof ( buf ) );
		if ( rc > SF_INITIATOR_LEN_MAX )
			return -SC_BAD_ITOR_IQN;
	}

	/* Check that target dhcp flag has value 0 or 1. */
	dhcp = sf_sett_get_intz ( settings, &target_dhcp_setting );
	if ( dhcp != 0 && dhcp != 1 )
		return -SC_BAD_TGT_DHCP_FLAG;

	/* Check target parameters only if value of initiator dhcp flag
	 * is zerro.
	 */
	if ( !dhcp ) {
		unsigned long lun, port;
		char buf[256];

		/* Check that the server name exists and is not too long */
		rc = fetch_string_setting ( settings, &server_name_setting,
					    buf, sizeof ( buf ) );
		if ( rc < 0 || rc > SF_TARGET_SERVER_LEN_MAX )
			return -SC_BAD_TGT_NAME;

		/* Check that target name is exists and not too long. */
		rc = fetch_string_setting ( settings, &target_name_setting,
					    buf, sizeof ( buf ) );
		if ( rc < 0 || rc > SF_TARGET_IQN_LEN_MAX )
			return -SC_BAD_TGT_IQN;

		/* Check that lun value is not too big. */
		( void ) fetch_uint_setting ( settings, &lun_setting, &lun );
		if ( lun > 0xff )
			return -SC_BAD_LUN;
		/* Check that target port value is correct. */
		( void ) fetch_uint_setting ( settings, &port_setting, &port );
		if ( port == 0 || port > 0xffff )
			return -SC_BAD_PORT;
	}

	if ( (rc = sf_check_chap(settings, &chap_setting,
	                         &username_setting,
	                         &password_setting)) < 0 ) {
		return rc;
	}
	if ( (rc = sf_check_chap(settings, &chap_target_setting,
	                         &reverse_username_setting,
	                         &reverse_password_setting)) < 0 ) {
		return rc;
	}

	/* Check that initiator IQN dhcp flag setting has value 0 or 1. */
	var = sf_sett_get_intz ( settings, &initiator_iqn_dhcp_setting );
	if ( var != 0 && var != 1 )
		return -SC_BAD_IQN_DHCP_FLAG;

	return 0;
}
#endif

/* NVO sanity checks for booting from net device. */
int
sf_nvo_checks_netdev(struct net_device *dev)
{
	struct settings  	*settings = NULL;
	long             	boot_type;
	int rc = 0;

	if ( (settings = netdev_nvo_settings(dev)) == NULL)
		return -SC_NO_NVO;
	boot_type = sf_sett_get_intz(settings, &boot_type_setting);

	switch (boot_type) {
	case BOOTTYPE_DISABLED:
		return 0;

	case BOOTTYPE_PXE:
		if ((rc = checks_pxe(dev)) != 0)
			return rc;
		break;

#ifdef BCV_PROTO_ISCSI
	case BOOTTYPE_ISCSI:
		if ((rc = checks_iscsi_single(dev)) != 0)
			return rc;
		break;
#endif

	default:
		return -SC_BAD_BOOTTYPE;
	}

	/* Checks which would not fit anywhere above would go here */
	return 0;
}
