/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdio.h>
#include <console.h>
#include <string.h>
#include <stdlib.h>
#include "sf_brwidgets.h"

#define MIN(a, b) ((a) < (b) ? (a) : (b))

int br_row;

/**
 * Prepare a widget list by an array of brief descriptions.
 * You can call prepare_conf_widgets() multiple times with the
 * same window
 *
 * @v window		Window to add items to
 * @v settings		Core settings structures
 * @v layout		Brief widgets descriptions
 * @v layout_size	Widgets quantity
 * @v option_list	Data for 'List' widgets
 * @v aud_data		Additional data
 *
 * @ret			Pointer to widgets list
 */
sf_widget * prepare_conf_widgets ( sf_widget *window, int margin,
				   struct settings **settings,
				   const struct sf_widget_brief *layout,
				   int layout_size, const sf_option *option_list,
				   void *aux_data ) {
	int			i, width;
	sf_widget		*new_widget;
	sf_widget		*list = NULL;
	struct settings		*layout_settings;

	/* If the user is creating a new window, then reset the position */
	if ( window == NULL )
		br_row = SFC_INIT_ROW;

	for ( i = 0; i < layout_size; i++ )
	{
		if ( layout[i].type == SFW_NEWLINE ) {
			br_row += layout[i].height;
			continue;
		}

		layout_settings = settings ? settings[layout[i].settings] : NULL;

		new_widget = NULL;
		switch ( layout[i].type ) {
		case SFW_WINDOW:
			new_widget = sf_window_init (
				layout[i].text, layout[i].top, layout[i].left,
				layout[i].width, layout[i].height );

			if ( window == NULL ) {
				window = new_widget;
				br_row = SFC_INIT_ROW;
				if ( layout[i].left == 0 && layout[i].text != NULL )
					++br_row;
			}
			break;

		case SFW_CHECKBOX:
			new_widget = sf_checkbox_init (
				window->wnd, br_row, margin, layout[i].text,
				layout_settings, layout[i].setting );
			++br_row;
			break;

		case SFW_EDITBOX:
			/* Cope with truncation */
			width = layout[i].width ? layout[i].width : 0x100;

			new_widget = sf_editbox_init (
				window->wnd, br_row, margin, width,
				layout[i].text, layout_settings, layout[i].setting,
				MIN ( width, SFC_EDITBOX_WIDTH ) );
			if ( new_widget != NULL &&
			     layout[i].top != '\0' ) {
				sf_editbox *box = SF_EDITBOX ( new_widget );
				box->shadow = layout[i].top;
			}

			++br_row;
			break;

		case SFW_BUTTON:
			new_widget = sf_button_init (
				layout[i].text, window->wnd, br_row,
				layout[i].left );
			br_row += layout[i].height;
			break;

		case SFW_COMBO:
			new_widget = sf_combo_init (
				window->wnd, br_row, margin, layout[i].text,
				layout_settings, layout[i].setting, option_list,
				layout[i].height );
			++br_row;
			option_list += layout[i].height;
			break;

		case SFW_TEXT:
			new_widget = sf_text_init (
				window->wnd, br_row, margin, layout[i].text );
			++br_row;
			break;

		case SFW_INFO:
			new_widget = sf_info_init (
				window->wnd, br_row, margin, layout[i].text, NULL );
			++br_row;
			break;

		case SFW_LIST:
			new_widget = sf_list_init (
				window->wnd, br_row, layout[i].left,
				window->wnd->width - 2 * layout[i].left,
				layout[i].height );
			br_row += layout[i].height;
			break;

		case SFW_ROW:
			new_widget = sf_row_init (
				window->wnd, margin, layout[i].text,
				layout_settings, layout[i].setting );
			break;

		case SFW_CAPTION:
			new_widget = sf_text_init (
				window->wnd, br_row, layout[i].left,
				layout[i].text );
			br_row += layout[i].height;
			break;
		default:
			continue;
		}

		if ( new_widget == NULL ) {
			DBG ( "widget type %d not created", layout[i].type );

			sf_widgets_list_free ( list );
			return NULL;
		}

		new_widget->action = layout[i].action;
		new_widget->aux_data = aux_data;
		new_widget->aux_data2 = layout[i].aux_data2;
		new_widget->help = layout[i].help;

		/* Load the value out of settings */
		sf_widget_load ( new_widget );

		if ( list == NULL )
			list = new_widget;
		else
			sf_widget_attach ( list, new_widget );
	}

	/* Add the list of new controls to the parent window. sf_widget_attach()
	 * copes with the distinct possibility the two lists are the same */
	if ( window != NULL && list != NULL )
		sf_widget_attach ( window, list );

	return list;
}

